#!/usr/bin/env python2
# -*- coding: utf-8 -*-

#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from timmy.analyze_health import GREEN, UNKNOWN, YELLOW, RED
from timmy.env import project_name
import logging
import re
import yaml


logger = logging.getLogger(project_name)


def register(function_mapping):
    function_mapping['rabbitmqctl-list-queues'] = parse_list_queues
    function_mapping['rabbitmqctl-status'] = parse_status


def parse_list_queues(stdout, script, node, stderr=None, exitcode=None):
    warning = 100
    error = 1000
    health = GREEN
    details = []
    if exitcode:
        health = UNKNOWN
        return health, details
    data = [l.rstrip() for l in stdout.splitlines()]
    for line in data[1:]:
        elements = line.rstrip().split()
        if len(elements) < 2:
            logger.warning('no value in list_queues: "%s"' % line.rstrip())
        else:
            count = int(elements[1])
            if count < error and count >= warning:
                health = max(health, YELLOW)
                details.append(line)
    return health, details


def prepare_status(stdout):
    bad_yaml = ''.join(stdout.splitlines()[1:])
    # quoting string elements
    bad_yaml = re.sub(r'([,{])([a-z_A-Z]+)([,}])', r'\1"\2"\3', bad_yaml)
    # changing first element int a key - replacing , with :
    bad_yaml = re.sub(r'({[^,]+),', r'\1:', bad_yaml)
    bad_yaml_list = list(bad_yaml)
    good_yaml, _ = fix_dicts(bad_yaml_list, 0)
    status_list = yaml.load(''.join(good_yaml))
    status_dict = squash_dicts(status_list)
    return status_dict


def fix_dicts(json_str_list, pos):
    '''recursively puts all comma-separted values into square
    brackets to make data look like normal 'key: value' dicts
    '''
    quoted_string = False
    value = True
    value_pos = 0
    commas = False
    is_list = False
    in_list = 0
    while pos < len(json_str_list):
        if not quoted_string:
            if json_str_list[pos] == '{':
                json_str_list, pos = fix_dicts(json_str_list, pos+1)
            elif json_str_list[pos] == '"':
                quoted_string = True
            elif json_str_list[pos] == ':':
                value = True
                value_pos = pos + 1
            elif json_str_list[pos] == '[':
                if value and not commas:
                    is_list = True
                    in_list += 1
            elif json_str_list[pos] == ']':
                in_list -= 1
            elif json_str_list[pos] == ',':
                commas = True
                if not in_list:
                    is_list = False
            elif json_str_list[pos] == '}':
                if not is_list and commas:
                    json_str_list = (json_str_list[:value_pos] + ['['] +
                                     json_str_list[value_pos:pos] + [']'] +
                                     json_str_list[pos:])
                pos += 2
                return json_str_list, pos
        elif json_str_list[pos] == '"':
            quoted_string = False
        pos += 1
    return json_str_list, pos


def squash_dicts(input_data):
    # recursively converts [{a:1},{b:2},{c:3}...] into {a:1, b:2, c:3}'''
    if type(input_data) is list:
        for i in range(len(input_data)):
            input_data[i] = squash_dicts(input_data[i])
        if all([type(i) is dict for i in input_data]):
            kv_list = [(k, v) for i in input_data for k, v in i.items()]
            input_data = dict(kv_list)
    elif type(input_data) is dict:
        for k, v in input_data.items():
            input_data[k] = squash_dicts(v)
    return input_data


def parse_status(stdout, script, node, stderr=None, exitcode=None):
    health = GREEN
    details = []
    status = prepare_status(stdout)
    if not status:
        health = UNKNOWN
        details = ['Status unavailable']
        if exitcode:
            if exitcode == 69:
                health = RED
                details = ['RabbitMQ is not running']
        return health, details

    # disk free check
    try:
        dfree = int(status['disk_free'])
        dlimit = int(status['disk_free_limit'])
        dfree_ok = 10**9  # 1GB
        if dfree > dlimit and dfree < dfree_ok:
            health = max(health, YELLOW)
            details.append('disk_free: %s, disk_free_limit: %s'
                           % (dfree, dlimit))
        elif dfree <= dlimit:
            health = max(health, RED)
            details.append('disk_free: %s, disk_free_limit: %s'
                           % (dfree, dlimit))
    except ValueError:
        details.append('cannot convert disk_free* to int')
        health = max(health, UNKNOWN)
    except KeyError:
        details.append('disk_free* not present')
        health = max(health, UNKNOWN)

    # process limit check
    try:
        pused = float(status['processes']['used'])
        plimit = float(status['processes']['limit'])
        ok_ratio = 0.9
        if pused < plimit and pused/plimit > ok_ratio:
            health = max(health, YELLOW)
            details.append('processes used: %s, processes limit: %s'
                           % (pused, plimit))
        elif pused >= plimit:
            health = max(health, RED)
            details.append('processes used: %s, processes limit: %s'
                           % (pused, plimit))
    except ValueError:
        details.append('cannot convert processes* to numbers')
        health = max(health, UNKNOWN)
    except KeyError:
        details.append('processes* not present')
        health = max(health, UNKNOWN)

    # fd check
    try:
        sused = float(status['file_descriptors']['sockets_used'])
        slimit = float(status['file_descriptors']['sockets_limit'])
        ok_ratio = 0.9
        if sused < slimit and sused/slimit > ok_ratio:
            health = max(health, YELLOW)
            details.append('sockets used: %s, sockets limit: %s'
                           % (sused, slimit))
        elif sused >= slimit:
            health = max(health, RED)
            details.append('sockets used: %s, sockets limit: %s'
                           % (sused, slimit))
        fdused = float(status['file_descriptors']['total_used'])
        fdlimit = float(status['file_descriptors']['total_limit'])
        ok_ratio = 0.9
        if fdused < fdlimit and fdused/fdlimit > ok_ratio:
            health = max(health, YELLOW)
            details.append('fd total used: %s, fd total limit: %s'
                           % (fdused, fdlimit))
        elif fdused >= fdlimit:
            health = max(health, RED)
            details.append('fd total used: %s, fd total limit: %s'
                           % (fdused, fdlimit))
    except ValueError:
        details.append('cannot convert file_descriptors* to numbers')
        health = max(health, UNKNOWN)
    except KeyError:
        details.append('file_descriptors* not present')
        health = max(health, UNKNOWN)

    return health, details



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import os
import sys
import timmy
from timmy.analyze_health import GREEN, UNKNOWN, YELLOW, RED
from timmy.env import project_name


logger = logging.getLogger(project_name)


def analyze(node_manager):
    def is_module(f):
        return f.endswith('.py') and not f.startswith('__')

    fn_mapping = {}
    modules_dir = 'analyze_modules'
    scan_path = os.path.join(os.path.dirname(__file__), modules_dir)
    base_path = os.path.split(timmy.__path__[0])[0]
    for item in os.walk(scan_path):
        for module_path in [m for m in item[2] if is_module(m)]:
            module_full_path = os.path.join(scan_path, module_path)
            module_rel_path = os.path.relpath(module_full_path, base_path)
            module_rel_path_noext = os.path.splitext(module_rel_path)[0]
            module_name = module_rel_path_noext.replace(os.path.sep, '.')
            module = __import__(module_name, fromlist=[project_name])
            module.register(fn_mapping)

    results = {}
    for node in node_manager.sorted_nodes():
        if not node.mapscr:
            node.generate_mapscr()
        for script, param in node.mapscr.items():
            if script in fn_mapping:
                if not os.path.exists(param['output_path']):
                    logger.warning('File %s does not exist'
                                   % param['output_path'])
                    continue
                try:
                    with open(param['output_path'], 'r') as f:
                        stdout = f.read()
                    if os.path.exists(param['stderr_path']):
                        with open(param['stderr_path'], 'r') as f:
                            stderr = f.read()
                            try:
                                exitcode = stderr.splitlines()[0]
                                exitcode = int(exitcode.rstrip().split()[1])
                            except IndexError:
                                logger.warning('Could not extract exitcode'
                                               ' from file "%s"' %
                                               param['stderr_path'])
                            except ValueError:
                                logger.warning('Could not convert exitcode'
                                               ' in file "%s" to int' %
                                               param['stderr_path'])
                    else:
                        stderr = None
                        exitcode = 0
                except IOError as e:
                    logger.warning('Could not read from %s' % e.filename)
                health, details = fn_mapping[script](stdout, script, node,
                                                     stderr=stderr,
                                                     exitcode=exitcode)
                if node.repr not in results:
                    results[node.repr] = []
                results[node.repr].append({'script': script,
                                           'output_file': param['output_path'],
                                           'stderr_file': param['stderr_path'],
                                           'exitcode': exitcode,
                                           'health': health,
                                           'details': details})
    node_manager.analyze_results = results


def analyze_print_results(node_manager):
    code_colors = {GREEN: ['GREEN', '\033[92m'],
                   UNKNOWN: ['UNKNOWN', '\033[94m'],
                   YELLOW: ['YELLOW', '\033[93m'],
                   RED: ['RED', '\033[91m']}
    color_end = '\033[0m'
    print('Nodes health analysis:')
    ag = True
    for node, result in sorted(node_manager.analyze_results.items()):
        node_health = max([x['health'] for x in result])
        node_color = code_colors[node_health][1]
        health_repr = code_colors[node_health][0]
        if node_health == 0:
            continue
        ag = False
        print('    %s%s: %s%s' % (node_color, node, health_repr, color_end))
        for r in result:
            if r['health'] == 0:
                continue
            color = code_colors[r['health']][1]
            sys.stdout.write(color)
            health_repr = code_colors[r['health']][0]
            print('%s%s: %s' % (8*' ', r['script'], health_repr))
            print('%s%s: %s' % (12*' ', 'output_file', r['output_file']))
            if ((r['stderr_file'] is not None) and
                    (os.path.exists(r['stderr_file']))):
                print('%s%s: %s' % (12*' ', 'stderr_file', r['stderr_file']))
            if r['exitcode'] != 0:
                print('%s%s: %s' % (12*' ', 'exitcode', r['exitcode']))
            if len(r['details']) > 1:
                print('%sdetails:' % (12*' '))
                for d in r['details']:
                    print('%s- %s' % (16*' ', d))
            elif r['details']:
                print('%sdetails: %s' % (12*' ', r['details'][0]))
            sys.stdout.write(color_end)
    if ag:
        sys.stdout.write(code_colors[GREEN][1])
        print('All green')
        sys.stdout.write(color_end)
